/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla XForms support.
 *
 * The Initial Developer of the Original Code is
 * IBM Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Aaron Reed <aaronr@us.ibm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef nsXFormsXPathEvaluator_h__
#define nsXFormsXPathEvaluator_h__

#include "nsIXFormsXPathEvaluator.h"
#include "txIXPathContext.h"
#include "nsIXPathEvaluatorInternal.h"
#include "nsIWeakReference.h"
#include "nsAutoPtr.h"
#include "txResultRecycler.h"


class nsIDOMDocument;
class nsIDOMXPathExpression;

/**
 * A class for evaluating an XPath expression string
 */
class nsXFormsXPathEvaluator : public nsIXFormsXPathEvaluator
{
public:
    nsXFormsXPathEvaluator();
    virtual ~nsXFormsXPathEvaluator();

    // nsISupports interface
    NS_DECL_ISUPPORTS

    // nsIXFormsXPathEvaluator interface
    NS_DECL_NSIXFORMXPATHEVALUATOR

private:
    // txIParseContext implementation
    class XFormsParseContextImpl : public txIParseContext
    {
    public:
        XFormsParseContextImpl(nsIDOMNode* aResolverNode, nsIDOMNode* aOrigCtxt)
              : mResolverNode(aResolverNode),
                mOriginalContext(aOrigCtxt),
                mLastError(NS_OK)
        {
        }

        ~XFormsParseContextImpl()
        {
        }

        nsresult getError()
        {
            return mLastError;
        }

        nsresult resolveNamespacePrefix(nsIAtom* aPrefix, PRInt32& aID);
        nsresult resolveFunctionCall(nsIAtom* aName, PRInt32 aID,
                                     FunctionCall*& aFunction);
        PRBool caseInsensitiveNameTests();
        void SetErrorOffset(PRUint32 aOffset);

    private:
        nsIDOMNode* mResolverNode;
        nsIDOMNode* mOriginalContext;
        nsresult mLastError;
    };

    nsRefPtr<txResultRecycler> mRecycler;
};

#endif
